"""Matching constraints for the matching metric."""
from enum import Enum, auto


class MatchingConstraint(Enum):
    """Matching constraints for the matching metric."""

    ONE_TO_ONE = auto()
    ONE_TO_MANY = auto()
    MANY_TO_ONE = auto()
    MANY_TO_MANY = auto()

    @staticmethod
    def from_str(s: str) -> "MatchingConstraint":
        return {
            "<->": MatchingConstraint.ONE_TO_ONE,
            "<-": MatchingConstraint.ONE_TO_MANY,
            "->": MatchingConstraint.MANY_TO_ONE,
            "~": MatchingConstraint.MANY_TO_MANY,
            "1:1": MatchingConstraint.ONE_TO_ONE,
            "1:*": MatchingConstraint.ONE_TO_MANY,
            "*:1": MatchingConstraint.MANY_TO_ONE,
            "*:*": MatchingConstraint.MANY_TO_MANY,
        }[s]



"""Metric suite definitions for coreference resolution."""
from typing import Collection

import metametric.core.dsl as mm
from metametric.core.metric import Metric
from metametric.structures.ie import Entity, EntitySet, Membership, Mention


def _muc_common_links(x: Entity, y: Entity) -> int:
    return max(0, len(set(x.mentions) & set(y.mentions)) - 1)


muc_link: Metric[Entity] = mm.from_func(_muc_common_links)

muc = mm.dataclass[EntitySet]({
    "entities": mm.set_matching[Entity, "~"](muc_link)
})


def _entity_set_to_membership_set(es: EntitySet) -> Collection[Membership]:
    return [Membership(mention=m, entity=e) for e in es.entities for m in e.mentions]


b_cubed_precision = mm.preprocess(
    _entity_set_to_membership_set,
    mm.set_matching[Membership, "<->", "precision"](
        mm.dataclass[Membership]({
            "mention": ...,
            "entity": mm.normalize["precision"](mm.auto[Entity])
        })
    )
)

b_cubed_recall = mm.preprocess(
    _entity_set_to_membership_set,
    mm.set_matching[Membership, "<->", "recall"](
        mm.dataclass[Membership]({
            "mention": ...,
            "entity": mm.normalize["recall"](mm.auto[Entity])
        })
    )
)


ceaf_phi4 = mm.dataclass[EntitySet]({
    "entities": mm.set_matching[Entity, "<->"](
        mm.dataclass[Entity]({
            "mentions": mm.set_matching[Mention, "<->", "f1"](...)
        })
    )
})


coref_suite = mm.suite({
    "muc": mm.family(muc, mm.macro_average(["precision", "recall", "f1"])),
    "b_cubed": mm.suite({
            "precision": mm.family(b_cubed_precision, mm.macro_average(["none"])),
            "recall": mm.family(b_cubed_recall, mm.macro_average(["none"])),
        }).with_extra(lambda m: {
            "f1": (
                2 * m["precision"] * m["recall"] / (m["precision"] + m["recall"])
                if (m["precision"] + m["recall"]) > 0 else 0.0
            )
        }),
    "ceaf_phi4": mm.family(ceaf_phi4, mm.macro_average(["precision", "recall", "f1"])),
}).with_extra(lambda m: {
    "avg-f1": (m["muc-f1"] + m["b_cubed-f1"] + m["ceaf_phi4-f1"]) / 3
})



"""Metric suites are collections of metrics that are computed together."""
from dataclasses import dataclass
from functools import cached_property
from typing import Callable, Dict, Protocol, Sequence, TypeVar

from metametric.core.metric import Metric
from metametric.core.reduction import Reduction
from metametric.core.state import (MetricState, MultipleMetricStates,
                                   SingleMetricState)

T = TypeVar("T", contravariant=True)


class Aggregator(Protocol[T]):
    """Metric aggregator for computing metrics on a stream of predictions and references."""

    @property
    def state(self) -> MetricState[T]:
        """The internal state of the aggregator."""
        raise NotImplementedError()

    def update_single(self, pred: T, ref: T) -> None:
        """Update the aggregator with a single prediction and its reference."""
        self.state.update_single(pred, ref)

    def update_batch(self, preds: Sequence[T], refs: Sequence[T]):
        """Update the aggregator with a batch of predictions and their references."""
        self.state.update_batch(preds, refs)

    def reset(self) -> None:
        """Reset the aggregator to its initialization state."""
        self.state.reset()

    def compute(self) -> Dict[str, float]:
        """Compute the metrics from the aggregator."""
        raise NotImplementedError()


class MetricSuite(Protocol[T]):
    """Metric suites are collections of metrics that are computed together."""
    def new(self) -> Aggregator[T]:
        """Create a new aggregator for the metric suite."""
        raise NotImplementedError()

    def with_extra(self, extra: Callable[[Dict[str, float]], Dict[str, float]]) -> "MetricSuite[T]":
        """Add extra metrics to the metric suite."""
        return MetricSuiteWithExtra(self, extra)


@dataclass
class MetricFamily(MetricSuite[T]):
    """A collection of metrics that shared the same internal state.

    For example, the precision, recall, and F-1 of event detection should be computed together within one family.
    """
    def __init__(self, metric: Metric[T], reduction: Reduction):
        self.metric = metric
        self.reduction = reduction

    def new(self) -> Aggregator[T]:
        return MetricFamilyAggregator(self)


class MetricFamilyAggregator(Aggregator[T]):
    """Aggregator for a metric family."""
    def __init__(self, family: MetricFamily[T]):
        self.family = family
        self._state = SingleMetricState(family.metric)

    @property
    def state(self) -> MetricState[T]:
        return self._state

    def compute(self) -> Dict[str, float]:
        return self.family.reduction.compute(self._state)


class MultipleMetricFamilies(MetricSuite[T]):
    """A collection of metric families, whose internal states are separate."""
    def __init__(
            self,
            families: Dict[str, MetricSuite[T]]
    ):
        self.families = families

    def new(self) -> Aggregator[T]:
        return MultipleMetricFamiliesAggregator(self)


class MultipleMetricFamiliesAggregator(Aggregator[T]):
    """Aggregator for multiple metric families."""
    def __init__(self, coll: MultipleMetricFamilies[T]):
        self.aggs: Dict[str, Aggregator[T]] = {
            name: family.new()
            for name, family in coll.families.items()
        }

    @cached_property
    def state(self) -> MetricState[T]:
        return MultipleMetricStates({
            name: agg.state
            for name, agg in self.aggs.items()
        })

    def compute(self) -> Dict[str, float]:
        metrics = {
            f"{name}-{key}" if key != "" else name: value
            for name, agg in self.aggs.items()
            for key, value in agg.compute().items()
        }
        return metrics


class MetricSuiteWithExtra(MetricSuite[T]):
    """A metric suite with extra metrics."""
    def __init__(self, original: MetricSuite[T], extra: Callable[[Dict[str, float]], Dict[str, float]]):
        self.original = original
        self.extra = extra

    def new(self) -> Aggregator[T]:
        return WithExtraAggregator(self)


class WithExtraAggregator(Aggregator[T]):
    """Aggregator for a metric suite with extra metrics."""
    def __init__(self, coll: MetricSuiteWithExtra[T]):
        self.agg = coll.original.new()
        self.extra = coll.extra

    @property
    def state(self) -> MetricState[T]:
        return self.agg.state

    def compute(self) -> Dict[str, float]:
        metrics = self.agg.compute()
        metrics.update(self.extra(metrics))
        return metrics



"""Abstract Meaning Representation (AMR) structure."""
from dataclasses import dataclass
from typing import Collection, Union

from metametric.core.decorator import metametric
from metametric.core.metric import Variable


@metametric()
@dataclass(eq=True, frozen=True)
class Prop:
    """A Proposition in an AMR."""

    subj: Variable
    pred: str
    obj: Union[Variable, str]


@metametric()
@dataclass
class AMR:
    """Abstract Meaning Representation (AMR) structure."""

    props: Collection[Prop]



"""This module contains the domain-specific language (DSL) for defining metrics."""
import sys
from dataclasses import dataclass, fields, is_dataclass
from typing import (
    Callable,
    Collection,
    Dict,
    Generic,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    get_origin,
    TYPE_CHECKING,
)

if sys.version_info >= (3, 10):
    from types import EllipsisType as Ell
else:
    if TYPE_CHECKING:
        from builtins import ellipsis as Ell
    else:
        Ell = type(...)

from metametric.core.constraint import MatchingConstraint
from metametric.core.matching import (
    GraphMatchingMetric,
    LatentSetMatchingMetric,
    SequenceMatchingMetric,
    SetMatchingMetric,
)
from metametric.core.decorator import derive_metric
from metametric.core.graph import Graph
from metametric.core.metric import ContramappedMetric, DiscreteMetric, Metric, ProductMetric, UnionMetric
from metametric.core.metric_suite import MetricFamily, MetricSuite, MultipleMetricFamilies
from metametric.core.normalizers import NormalizedMetric, Normalizer
from metametric.core.reduction import MacroAverage, MicroAverage, MultipleReductions, Reduction

T = TypeVar("T", contravariant=True)
S = TypeVar("S")

DslConfig = Union[
    Type[T],
    Tuple[Type[T], Union[MatchingConstraint, str]],
    Tuple[Type[T], Union[MatchingConstraint, str], Union[Normalizer, str, None]],
]


@dataclass
class _Config(Generic[T]):
    cls: Type[T]
    constraint: MatchingConstraint = MatchingConstraint.ONE_TO_ONE
    normalizer: Optional[Normalizer] = None

    @classmethod
    def standardize(cls, config: DslConfig[T]) -> "_Config[T]":
        if isinstance(config, tuple):
            if len(config) == 2:
                t, constraint = config
                normalizer = None
            elif len(config) == 3:
                t, constraint, normalizer = config
                if isinstance(normalizer, str):
                    normalizer = Normalizer.from_str(normalizer)
            else:
                raise ValueError(f"Invalid config: {config}")
        else:
            t = config
            constraint = MatchingConstraint.ONE_TO_ONE
            normalizer = None
        if isinstance(constraint, str):
            constraint = MatchingConstraint.from_str(constraint)
        return cls(t, constraint, normalizer)


def from_func(func: Callable[[T, T], float]) -> Metric[T]:
    """Create a metric from a binary function."""
    return Metric.from_function(func)


def preprocess(func: Callable[[S], T], m: Metric[T]) -> Metric[S]:
    """Preprocess the input by some function then apply a metric.

    This is the `contramap` operation on the metric functor.
    """
    return ContramappedMetric(m, func)


class _Auto:
    def __getitem__(self, config: DslConfig[T]) -> Metric[T]:
        cfg = _Config.standardize(config)
        return derive_metric(cfg.cls, constraint=cfg.constraint)


auto = _Auto()


class _Discrete:
    def __getitem__(self, t: Type[T]) -> Metric[T]:
        return DiscreteMetric(t)


discrete = _Discrete()


class _DataClass:
    def __getitem__(self, config: DslConfig[T]) -> Callable[[Dict[str, Union[Ell, Metric]]], Metric[T]]:
        cfg = _Config.standardize(config)
        assert is_dataclass(cfg.cls)

        def product_metric(field_metrics: Dict[str, Union[Ell, Metric]]) -> Metric[T]:
            field_types = {fld.name: fld.type for fld in fields(cfg.cls)}
            field_metrics_no_ell: Dict[str, Metric] = {
                fld: (auto[field_types[fld], cfg.constraint] if metric is ... else metric)
                for fld, metric in field_metrics.items()
            }
            return ProductMetric(cls=cfg.cls, field_metrics=field_metrics_no_ell)

        return product_metric


dataclass = _DataClass()


class _Union:
    def __getitem__(self, config: DslConfig[T]) -> Callable[[Dict[Type, Union[Ell, Metric]]], Metric[T]]:
        cfg = _Config.standardize(config)

        def union_metric(case_metrics: Dict[Type, Union[Ell, Metric]]) -> Metric[T]:
            assert get_origin(cfg.cls) is Union
            case_metrics_no_ell: Dict[type, Metric] = {
                case: (auto[case, cfg.constraint] if metric is ... else metric)
                for case, metric in case_metrics.items()
            }
            return UnionMetric(cls=cfg.cls, case_metrics=case_metrics_no_ell)

        return union_metric


union = _Union()


class _SetMatching:
    def __getitem__(self, config: DslConfig[T]) -> Callable[[Union[Ell, Metric[T]]], Metric[Collection[T]]]:
        cfg = _Config.standardize(config)

        def matching_metric(inner: Union[Ell, Metric[T]]) -> Metric[Collection[T]]:
            if inner is ...:
                inner = auto[cfg.cls, cfg.constraint]
            match = SetMatchingMetric(inner, constraint=cfg.constraint)
            if cfg.normalizer is not None:
                match = NormalizedMetric(match, cfg.normalizer)
            return match

        return matching_metric


set_matching = _SetMatching()


class _SequenceMatching:
    def __getitem__(self, config: DslConfig[T]) -> Callable[[Union[Ell, Metric[T]]], Metric[Sequence[T]]]:
        cfg = _Config.standardize(config)

        def matching_metric(inner: Union[Ell, Metric[T]]) -> Metric[Sequence[T]]:
            if inner is ...:
                inner = auto[cfg.cls, cfg.constraint]
            match = SequenceMatchingMetric(inner, constraint=cfg.constraint)
            if cfg.normalizer is not None:
                match = NormalizedMetric(match, cfg.normalizer)
            return match

        return matching_metric


sequence_matching = _SequenceMatching()


class _GraphMatching:
    def __getitem__(self, config: DslConfig[T]) -> Callable[[Union[Ell, Metric[T]]], Metric[Graph[T]]]:
        cfg = _Config.standardize(config)

        def matching_metric(inner: Union[Ell, Metric[T]]) -> Metric[Graph[T]]:
            if inner is ...:
                inner = auto[cfg.cls, cfg.constraint]
            match = GraphMatchingMetric(inner, constraint=cfg.constraint)
            if cfg.normalizer is not None:
                match = NormalizedMetric(match, cfg.normalizer)
            return match

        return matching_metric


graph_matching = _GraphMatching()


class _LatentSetMatching:
    def __getitem__(self, config: DslConfig[T]) -> Callable[[Union[Ell, Metric[T]]], Metric[Collection[T]]]:
        cfg = _Config.standardize(config)

        def latent_matching_metric(inner: Union[Ell, Metric[T]]) -> Metric[Collection[T]]:
            if inner is ...:
                inner = auto[cfg.cls, cfg.constraint]
            match = LatentSetMatchingMetric(cfg.cls, inner, constraint=cfg.constraint)
            if cfg.normalizer is not None:
                match = NormalizedMetric(match, cfg.normalizer)
            return match

        return latent_matching_metric


latent_set_matching = _LatentSetMatching()


class _Normalize:
    def __getitem__(self, normalizer: Union[Normalizer, str]) -> Callable[[Metric[T]], Metric[T]]:
        if isinstance(normalizer, str):
            normalizer_obj = Normalizer.from_str(normalizer)

        def normalize(metric: Metric[T]) -> Metric[T]:
            return metric if normalizer_obj is None else NormalizedMetric(metric, normalizer_obj)

        return normalize


normalize = _Normalize()


class _MacroAverage:
    def __call__(self, normalizers: Collection[Union[Normalizer, str]]) -> Reduction:
        normalizer_objs = [
            Normalizer.from_str(normalizer) if isinstance(normalizer, str) else normalizer
            for normalizer in normalizers
        ]
        return MacroAverage(normalizer_objs)


macro_average = _MacroAverage()


class _MicroAverage:
    def __call__(self, normalizers: Collection[Union[Normalizer, str]]) -> Reduction:
        normalizer_objs = [
            Normalizer.from_str(normalizer) if isinstance(normalizer, str) else normalizer
            for normalizer in normalizers
        ]
        return MicroAverage(normalizer_objs)


micro_average = _MicroAverage()


class _Family:
    def __call__(self, metric: Metric[T], reduction: Union[Reduction, Dict[str, Reduction]]) -> MetricFamily[T]:
        if isinstance(reduction, dict):
            reduction = MultipleReductions(reduction)
        return MetricFamily(metric, reduction)


family = _Family()


class _Suite:
    def __call__(self, collection: Dict[str, MetricSuite[T]]) -> MetricSuite[T]:
        return MultipleMetricFamilies(collection)


suite = _Suite()



from abc import ABC, abstractmethod
from dataclasses import dataclass, fields
from typing import (Any, Callable, Collection, Generic, Iterator, List,
                    Optional, Sequence, Type, TypeVar)

import numpy as np
import scipy as sp

from metametric.core.constraint import MatchingConstraint
from metametric.core.metric import Variable

T = TypeVar('T')


@dataclass
class ConstraintBuilder(ABC):
    n_x: int
    n_y: int
    n_x_vars: int
    n_y_vars: int

    def index_pair(self, i: int, j: int) -> int:
        return i * self.n_y + j

    def index_var_pair(self, i: int, j: int) -> int:
        return self.n_x * self.n_y + i * self.n_y_vars + j

    @abstractmethod
    def build(self) -> Optional[sp.optimize.LinearConstraint]:
        pass


@dataclass
class MatchingConstraintBuilder(ConstraintBuilder):
    constraint: MatchingConstraint = MatchingConstraint.ONE_TO_ONE

    def build(self) -> Optional[sp.optimize.LinearConstraint]:
        constraint_matrix_ctor: Callable[[int, int], Optional[np.ndarray]] = {
            MatchingConstraint.ONE_TO_ONE: _get_one_to_one_constraint_matrix,
            MatchingConstraint.ONE_TO_MANY: _get_one_to_many_constraint_matrix,
            MatchingConstraint.MANY_TO_ONE: _get_many_to_one_constraint_matrix,
            MatchingConstraint.MANY_TO_MANY: lambda _0, _1: None,
        }[self.constraint]
        m = constraint_matrix_ctor(self.n_x, self.n_y)
        if m is not None:
            m = np.concatenate(
                [
                    m,
                    np.zeros([m.shape[0], self.n_x_vars * self.n_y_vars]),
                ],   # pad with zeros for the latent variables
                axis=1,
            )
            return sp.optimize.LinearConstraint(
                A=m,
                ub=np.ones([m.shape[0]]),
            )
        else:
            return None  # no constraint when the constraint is MANY_TO_MANY


@dataclass
class VariableMatchingConstraintBuilder(ConstraintBuilder):

    def build(self) -> Optional[sp.optimize.LinearConstraint]:
        if self.n_x_vars == 0 or self.n_y_vars == 0:
            return None
        # only one-to-one matching between variables
        m = _get_one_to_one_constraint_matrix(self.n_x_vars, self.n_y_vars)
        m = np.concatenate(
            [
                np.zeros([m.shape[0], self.n_x * self.n_y]),
                m,
            ],  # pad with zeros for the matching items
            axis=1,
        )
        return sp.optimize.LinearConstraint(
            A=m,
            ub=np.ones([m.shape[0]]),
        )


@dataclass
class MonotonicityConstraintBuilder(ConstraintBuilder):
    gram_matrix: np.ndarray  # R[n_x, n_y]
    x_reachability: np.ndarray  # R[n_x, n_x]
    y_reachability: np.ndarray  # R[n_y, n_y]

    def build(self) -> Optional[sp.optimize.LinearConstraint]:
        vectors = []
        possible_matching_pairs = [
            (u, v)
            for u in range(self.n_x)
            for v in range(self.n_y)
            if self.gram_matrix[u, v] > 0
        ]
        for u0, v0 in possible_matching_pairs:
            for u1, v1 in possible_matching_pairs:
                if self.x_reachability[u0, u1] != self.y_reachability[v0, v1]:
                    vec = np.zeros(self.n_x * self.n_y)
                    vec[self.index_pair(u0, v0)] = 1
                    vec[self.index_pair(u1, v1)] = 1
                    vectors.append(vec)
                    # Enforce monotonicity of the matching
                    #    [u0 ~ v0] & [u1 ~ v1] -> [u0 <= u1] ≡ [v0 <= v1]
                    # => 1 - ((1 - t[u0~v0]) + (1 - t[u1~v1])) <= 1[[u0 <= u1] ≡ [v0 <= v1]]
                    # => t[u0~v0] + t[u1~v1] <= 1[[u0 <= u1] ≡ [v0 <= v1]] + 1
                    # => t[u0~v0] + t[u1~v1] <= 1  (if [u0 <= u1] ≡ [v0 <= v1], constraint redundant)

        if len(vectors) == 0:
            return None
        constraint_matrix = np.stack(vectors, axis=0)  # R[n_constraints, n_x * n_y]
        return sp.optimize.LinearConstraint(
            A=constraint_matrix,
            ub=np.ones([constraint_matrix.shape[0]]),
        )


@dataclass
class LatentVariableConstraintBuilder(ConstraintBuilder, Generic[T]):
    x: Collection[T]
    y: Collection[T]
    cls: Type[T]
    gram_matrix: np.ndarray  # R[n_x, n_y]

    def build(self) -> Optional[sp.optimize.LinearConstraint]:
        x_vars = list(_all_variables(self.x))
        y_vars = list(_all_variables(self.y))
        x_var_to_id = {t: i for i, t in enumerate(x_vars)}
        y_var_to_id = {t: j for j, t in enumerate(y_vars)}
        vectors = []
        for i, a in enumerate(self.x):
            for j, b in enumerate(self.y):
                if self.gram_matrix[i, j] > 0:
                    for fld in fields(self.cls):
                        a_fld = getattr(a, fld.name, None)
                        b_fld = getattr(b, fld.name, None)
                        if isinstance(a_fld, Variable) and isinstance(b_fld, Variable):
                            vec = np.zeros(self.n_x * self.n_y + self.n_x_vars * self.n_y_vars)
                            vec[self.index_pair(i, j)] = 1
                            vec[self.index_var_pair(x_var_to_id[a_fld], y_var_to_id[b_fld])] = -1
                            vectors.append(vec)
                            # Item matches implies variable matches
                            #    [a ~ b] -> [a_fld ~ b_fld]
                            # => t[a~b] <= t[a_fld~b_fld]
                            # => t[a~b] - t[a_fld~b_fld] <= 0
        if len(vectors) == 0:
            return None
        constraint_matrix = np.stack(vectors, axis=0)  # R[n_constraints, n_x * n_y + n_x_vars * n_y_vars]
        return sp.optimize.LinearConstraint(
            A=constraint_matrix,
            ub=np.zeros([constraint_matrix.shape[0]]),
        )


class MatchingProblem(Generic[T]):
    """Creates a matching problem that is solved by ILP.

    The constrained ILP problem has variables
    - for each pair of elements in X and Y, and
    - for each pair of potential latent variables in X and Y.
    """
    def __init__(
            self,
            x: Sequence[T],
            y: Sequence[T],
            gram_matrix: np.ndarray,
            has_vars: bool = False,
    ):
        self.x = x
        self.y = y
        self.n_x = len(x)
        self.n_y = len(y)
        self.gram_matrix = gram_matrix
        if has_vars:
            self.x_vars = list(_all_variables(x))
            self.y_vars = list(_all_variables(y))
            self.n_x_vars = len(self.x_vars)
            self.n_y_vars = len(self.y_vars)
        else:
            self.x_vars = []
            self.y_vars = []
            self.n_x_vars = 0
            self.n_y_vars = 0
        self.constraints: List[sp.optimize.LinearConstraint] = []

    def add_matching_constraint(self, constraint_type: MatchingConstraint):
        constraint = MatchingConstraintBuilder(
            n_x=self.n_x,
            n_y=self.n_y,
            n_x_vars=self.n_x_vars,
            n_y_vars=self.n_y_vars,
            constraint=constraint_type,
        ).build()
        if constraint is not None:
            self.constraints.append(constraint)

    def add_variable_matching_constraint(self):
        constraint = VariableMatchingConstraintBuilder(
            n_x=self.n_x,
            n_y=self.n_y,
            n_x_vars=self.n_x_vars,
            n_y_vars=self.n_y_vars,
        ).build()
        if constraint is not None:
            self.constraints.append(constraint)

    def add_monotonicity_constraint(self, x_reachability: np.ndarray, y_reachability: np.ndarray):
        constraint = MonotonicityConstraintBuilder(
            n_x=self.n_x,
            n_y=self.n_y,
            n_x_vars=self.n_x_vars,
            n_y_vars=self.n_y_vars,
            gram_matrix=self.gram_matrix,
            x_reachability=x_reachability,
            y_reachability=y_reachability,
        ).build()
        if constraint is not None:
            self.constraints.append(constraint)

    def add_latent_variable_constraint(self, cls: Type[T]):
        constraint = LatentVariableConstraintBuilder(
            n_x=self.n_x,
            n_y=self.n_y,
            n_x_vars=self.n_x_vars,
            n_y_vars=self.n_y_vars,
            x=self.x,
            y=self.y,
            cls=cls,
            gram_matrix=self.gram_matrix,
        ).build()
        if constraint is not None:
            self.constraints.append(constraint)

    def solve(self):
        # Layout of the constraint matrix:
        # ╒═════════════════════════╤═══════════════════════════╕
        # │        n_x * n_y        │    n_x_vars * n_y_vars    │
        # ┝━━━━━━━━━━━━━━━━━━━━━━━━━┿━━━━━━━━━━━━━━━━━━━━━━━━━━━┥
        # │   MATCHING_CONSTRAINT   │                           │
        # ├─────────────────────────┼───────────────────────────┤
        # │                         │  VAR_MATCHING_CONSTRAINT  │
        # ├─────────────────────────┼───────────────────────────┤
        # │ MONOTONICITY_CONSTRAINT │                           │
        # ├─────────────────────────┴───────────────────────────┤
        # │            LATENT_VARIABLE_CONSTRAINT               │
        # ╘═════════════════════════════════════════════════════┙
        coef = np.concatenate(
            [
                self.gram_matrix.reshape([self.n_x * self.n_y]),
                np.zeros([self.n_x_vars * self.n_y_vars]),
            ],  # pad with zeros for the latent variables
            axis=0,
        )
        result = sp.optimize.milp(
            c=-coef,
            constraints=self.constraints,
            bounds=sp.optimize.Bounds(lb=0, ub=1),
            integrality=np.ones_like(coef),
        )
        return -result.fun


def _get_one_to_many_constraint_matrix(n_x: int, n_y: int) -> np.ndarray:  # [Y, X * Y]
    mask_y = np.zeros([n_y, n_x, n_y])
    mask_y[np.arange(n_y), :, np.arange(n_y)] = 1
    mask_y = mask_y.reshape([n_y, n_x * n_y])
    return mask_y


def _get_many_to_one_constraint_matrix(n_x: int, n_y: int) -> np.ndarray:  # [X, X * Y]
    mask_x = np.zeros([n_x, n_x, n_y])
    mask_x[np.arange(n_x), np.arange(n_x), :] = 1
    mask_x = mask_x.reshape([n_x, n_x * n_y])
    return mask_x


def _get_one_to_one_constraint_matrix(n_x: int, n_y: int) -> np.ndarray:  # [X + Y, X * Y]
    mask_x = _get_one_to_many_constraint_matrix(n_x, n_y)
    mask_y = _get_many_to_one_constraint_matrix(n_x, n_y)
    return np.concatenate([mask_x, mask_y], axis=0)


def _all_variables(obj: Any) -> Iterator[Variable]:
    if isinstance(obj, Variable):
        yield obj
    elif isinstance(obj, Collection) and not isinstance(obj, str):
        for item in obj:
            yield from _all_variables(item)
    elif getattr(obj, "__dict__", None) is not None:
        for fld in vars(obj).values():
            yield from _all_variables(fld)



"""Defines the states of metric aggregators."""
from typing import Dict, Protocol, Sequence, TypeVar

from metametric.core.metric import Metric

T = TypeVar("T", contravariant=True)


class MetricState(Protocol[T]):
    """Encapsulates the state of a metric aggregator."""

    def update_single(self, pred: T, ref: T) -> None:
        """Update the aggregator with a single prediction and its reference."""
        raise NotImplementedError()

    def update_batch(self, preds: Sequence[T], refs: Sequence[T]) -> None:
        """Update the aggregator with a batch of predictions and their references."""
        for p, r in zip(preds, refs):
            self.update_single(p, r)

    def reset(self) -> None:
        """Reset the aggregator."""
        raise NotImplementedError()

    def __len__(self):
        """Return the number of predictions."""
        raise NotImplementedError()


class SingleMetricState(MetricState[T]):
    """Encapsulates the state of a single metric aggregator."""
    def __init__(self, metric: Metric[T]):
        self.metric = metric
        self.preds = []
        self.refs = []
        self.matches = []

    def update_single(self, pred: T, ref: T) -> None:
        """Update the aggregator with a single prediction and its reference."""
        sxx = self.metric.score_self(pred)
        syy = self.metric.score_self(ref)
        sxy = self.metric.score(pred, ref)
        self.preds.append(sxx)
        self.refs.append(syy)
        self.matches.append(sxy)

    def reset(self) -> None:
        """Reset the aggregator to its initialization state."""
        self.preds = []
        self.refs = []
        self.matches = []

    def __len__(self):
        """Returns the number of prediction-reference pairs aggregated."""
        return len(self.matches)


class MultipleMetricStates(MetricState[T]):
    """Encapsulates the state of multiple metric aggregators."""
    def __init__(self, states: Dict[str, MetricState[T]]):
        self.states = states

    def update_single(self, pred: T, ref: T) -> None:
        for state in self.states.values():
            state.update_single(pred, ref)

    def update_batch(self, preds: Sequence[T], refs: Sequence[T]) -> None:
        for state in self.states.values():
            state.update_batch(preds, refs)

    def reset(self) -> None:
        for state in self.states.values():
            state.reset()

    def __len__(self):
        """Returns the number of prediction-reference pairs aggregated."""
        return len(list(self.states.values())[0])



"""Decorator for deriving metrics from dataclasses."""
from dataclasses import fields, is_dataclass
from typing import (Annotated, Any, Callable, Collection, Literal, Type,
                    TypeVar, Union, get_args, get_origin)

from metametric.core.matching import (MatchingConstraint,
                                      LatentSetMatchingMetric,
                                      SetMatchingMetric)
from metametric.core.metric import (DiscreteMetric, HasLatentMetric, HasMetric,
                                    Metric, ProductMetric, UnionMetric,
                                    Variable)
from metametric.core.normalizers import NormalizedMetric, Normalizer

NormalizerLiteral = Literal["none", "jaccard", "dice", "f1"]
ConstraintLiteral = Literal["<->", "<-", "->", "~", "1:1", "1:*", "*:1", "*:*"]

T = TypeVar("T", contravariant=True)


def may_be_variable(cls: Any) -> bool:
    """Check if a type may be a `Variable`."""
    if cls is Variable:
        return True
    if get_origin(cls) is not None and get_origin(cls) is Union:
        if any(t is Variable for t in get_args(cls)):
            return True
    return False


def dataclass_has_variable(cls: Type) -> bool:
    """Check if a dataclass has a field is in `Variable` type."""
    if cls is Variable:
        return True
    if is_dataclass(cls):
        if any(may_be_variable(f.type) for f in fields(cls)):
            return True
    return False



def derive_metric(cls: Type, constraint: MatchingConstraint) -> Metric:
    """Derive a unified metric from any type.

    Args:
        cls (`Type`): The type to derive the metric from.
        constraint (`MatchingConstraint`): The matching constraint to use.

    Returns:
        `Metric`: The derived metric.
    """
    # if the type is annotated with a metric instance, use the metric annotation
    if get_origin(cls) is Annotated:
        metric = get_args(cls)[1]
        if isinstance(metric, Metric):
            return metric

    # if an explicit metric is defined, use it
    if isinstance(cls, HasMetric):
        return cls.metric

    cls_origin = get_origin(cls)
    if isinstance(cls, HasLatentMetric):
        return cls.latent_metric

    # derive product metric from dataclass
    elif is_dataclass(cls):
        return ProductMetric(
            cls=cls, field_metrics={fld.name: derive_metric(fld.type, constraint=constraint) for fld in fields(cls)}
        )

    # derive union metric from unions
    elif cls_origin is Union:
        return UnionMetric(
            cls=cls, case_metrics={case: derive_metric(case, constraint=constraint) for case in get_args(cls)}
        )
    # TODO: graph, dag, tree, sequence

    # derive set matching metric from collections
    elif cls_origin is not None and isinstance(cls_origin, type) and issubclass(cls_origin, Collection):
        elem_type = get_args(cls)[0]
        inner_metric = derive_metric(elem_type, constraint=constraint)
        if dataclass_has_variable(elem_type):
            return LatentSetMatchingMetric(
                cls=elem_type,
                inner=inner_metric,
                constraint=constraint,
            )
        else:
            return SetMatchingMetric(
                inner=inner_metric,
                constraint=constraint,
            )

    # derive discrete metric from equality
    elif getattr(cls, "__eq__", None) is not None:
        return DiscreteMetric(cls=cls)

    else:
        raise ValueError(f"Could not derive metric from type {cls}.")


def metametric(
    normalizer: Union[NormalizerLiteral, Normalizer] = "none",
    constraint: Union[ConstraintLiteral, MatchingConstraint] = "<->",
) -> Callable[[Type], Type]:
    """Decorate a dataclass to have corresponding metric derived.

    Args:
        normalizer (`Union[NormalizerLiteral, Normalizer]`, defaults to "none"):
            The normalizer to use.
        constraint (`ConstraintLiteral`, defaults to "<->"):
            The matching constraint to use.

    Returns:
        `Callable[[Type], Type]`: The class decorator.
    """

    def class_decorator(cls: Type) -> Type:
        nonlocal normalizer, constraint
        if isinstance(constraint, MatchingConstraint):
            metric = derive_metric(cls, constraint=constraint)
        else:
            metric = derive_metric(cls, constraint=MatchingConstraint.from_str(constraint))
        if isinstance(normalizer, Normalizer):
            normalized_metric = NormalizedMetric(metric, normalizer=normalizer)
        else:
            if normalizer == "none":
                normalized_metric = metric
            else:
                normalizer_obj = Normalizer.from_str(normalizer)
                assert normalizer_obj is not None
                normalized_metric = NormalizedMetric(metric, normalizer=normalizer_obj)
        if dataclass_has_variable(cls):
            setattr(cls, "latent_metric", normalized_metric)  # type: ignore
        else:
            setattr(cls, "metric", normalized_metric)  # type: ignore
        return cls

    return class_decorator



"""Define a `Graph` protocol that `networkx.DiGraph` satisfies."""
from typing import Collection, Iterator, Protocol, TypeVar, runtime_checkable

import numpy as np

T = TypeVar("T")


@runtime_checkable
class Graph(Protocol[T]):
    """A minimum graph definition of which `networkx.DiGraph` is an instance."""

    def nodes(self) -> Collection[T]:
        raise NotImplementedError()

    def successors(self, x: T) -> Iterator[T]:
        raise NotImplementedError()

    def predecessors(self, x: T) -> Iterator[T]:
        raise NotImplementedError()


def _adjacency_matrix(graph: Graph[T]) -> np.ndarray:
    """Get the adjacency matrix of a graph."""
    nodes = list(graph.nodes())
    node_to_id = {x: i for i, x in enumerate(nodes)}
    n = len(nodes)
    adj = np.zeros([n, n], dtype=bool)
    for i, x in enumerate(nodes):
        for y in graph.successors(x):
            j = node_to_id[y]
            adj[i, j] = 1
    return adj


def _reachability_matrix(graph: Graph[T]) -> np.ndarray:
    """Get the reachability matrix of a graph."""
    a = _adjacency_matrix(graph)
    b = np.eye(a.shape[0], dtype=bool) + a
    c = b @ b
    while not np.all(c == b):
        b = c
        c = b @ b
    return c




"""Data structures commonly used in information extraction.

The data structures defined here can automatically derive commonly used metrics in IE.
"""
from dataclasses import dataclass
from typing import Collection

from metametric.core.decorator import metametric


@metametric()
@dataclass(eq=True, frozen=True)
class Mention:
    """A mention span commonly used for .

    The mention span here does not enforce whether it is right exclusive or not.
    However, it is a convention that the mention span is right exclusive.

    Attributes:
    ----------
        left: The left index of the span.
        right: The right index of the span.
    """

    left: int
    right: int


@metametric()
@dataclass
class Relation:
    """A relation between two mentions commonly used in relation extraction.

    Attributes:
    ----------
        type: The type of the relation.
        subj: The subject mention.
        obj: The object mention.

    """

    type: str
    subj: Mention
    obj: Mention


@metametric()
class Trigger:
    """A trigger mention commonly used in event extraction.

    Attributes:
    ----------
        type: The type of the trigger, commonly used to indicate event type.
    """

    mention: Mention
    type: str


@metametric()
class Argument:
    """An argument mention commonly used in event extraction.

    Attributes:
    ----------
        mention: The mention of the argument.
        role: The role of the argument.
    """

    mention: Mention
    role: str


@metametric()
@dataclass
class Event:
    """An event commonly used in event extraction.

    Attributes:
    ----------
        trigger: The trigger of the event.
        args: The arguments of the event.
    """

    trigger: Trigger
    args: Collection[Argument]


@metametric()
@dataclass
class EventSet:
    """A set of events to present predicted or referenced events."""

    events: Collection[Event]


@metametric()
@dataclass
class RelationSet:
    """A set of relations to present predicted or referenced relations."""

    relations: Collection[Relation]


@dataclass
class Entity:
    """An entity comprises multiple mentions, commonly used in coreference resolution."""

    mentions: Collection[Mention]


@metametric()
@dataclass
class EntitySet:
    """A set of entities to present predicted or referenced entities."""

    entities: Collection[Entity]


@metametric()
@dataclass
class Membership:
    """A membership relation between an entity and a mention."""
    mention: Mention
    entity: Entity





"""Metric derivation with matching constraints."""
from dataclasses import is_dataclass
from typing import Collection, Sequence, Type, TypeVar, Union

import numpy as np
import scipy.optimize as spo

from metametric.core._ilp import MatchingProblem
from metametric.core.constraint import MatchingConstraint
from metametric.core.graph import Graph, _reachability_matrix
from metametric.core.metric import DiscreteMetric, Metric

T = TypeVar("T")


class SetMatchingMetric(Metric[Collection[T]]):
    """A metric derived from the matching of two sets."""

    def __init__(self, inner: Metric[T], constraint: Union[str, MatchingConstraint] = MatchingConstraint.ONE_TO_ONE):
        self.inner = inner
        self.constraint = MatchingConstraint.from_str(constraint) if isinstance(constraint, str) else constraint

    def score(self, x: Collection[T], y: Collection[T]) -> float:
        """Score two sets of objects."""
        if isinstance(self.inner, DiscreteMetric) and self.constraint == MatchingConstraint.ONE_TO_ONE:
            return len(set(x) & set(y))
        else:
            m = self.inner.gram_matrix(x, y)
            if self.constraint == MatchingConstraint.ONE_TO_ONE:
                row_idx, col_idx = spo.linear_sum_assignment(
                    cost_matrix=m,
                    maximize=True,
                )
                return m[row_idx, col_idx].sum()
            if self.constraint == MatchingConstraint.ONE_TO_MANY:
                return m.max(axis=0).sum()
            if self.constraint == MatchingConstraint.MANY_TO_ONE:
                return m.max(axis=1).sum()
            if self.constraint == MatchingConstraint.MANY_TO_MANY:
                return m.sum()
            raise ValueError(f"Invalid constraint: {self.constraint}")

    def score_self(self, x: Collection[T]) -> float:
        """Score a set of objects with itself."""
        if self.constraint == MatchingConstraint.MANY_TO_MANY:
            return self.inner.gram_matrix(x, x).sum()
        elif self.constraint == MatchingConstraint.ONE_TO_ONE:
            return sum(self.inner.score_self(u) for u in x)
        else:
            return self.score(x, x)


class SequenceMatchingMetric(Metric[Sequence[T]]):
    """A metric derived from the matching of two sequences."""

    def __init__(self, inner: Metric[T], constraint: Union[str, MatchingConstraint] = MatchingConstraint.ONE_TO_ONE):
        self.inner = inner
        self.constraint = MatchingConstraint.from_str(constraint) if isinstance(constraint, str) else constraint

    def score(self, x: Sequence[T], y: Sequence[T]) -> float:
        m = self.inner.gram_matrix(x, y)
        f = np.zeros([m.shape[0] + 1, m.shape[1] + 1])
        for i in range(m.shape[0] + 1):
            for j in range(m.shape[1] + 1):
                if i == 0 or j == 0:
                    f[i, j] = 0
                else:
                    f[i, j] = max(f[i - 1, j - 1] + m[i - 1, j - 1], f[i - 1, j], f[i, j - 1])
                    if self.constraint == MatchingConstraint.ONE_TO_MANY:
                        f[i, j] = max(f[i, j], f[i, j - 1] + m[i - 1, j - 1])
                    elif self.constraint == MatchingConstraint.MANY_TO_ONE:
                        f[i, j] = max(f[i, j], f[i - 1, j] + m[i - 1, j - 1])
                    elif self.constraint == MatchingConstraint.MANY_TO_MANY:
                        f[i, j] = max(f[i, j], f[i, j - 1] + m[i - 1, j - 1],
                                      f[i - 1, j] + m[i - 1, j - 1])
        return f[-1, -1].item()

    def score_self(self, x: Sequence[T]) -> float:
        if self.constraint == MatchingConstraint.ONE_TO_ONE:
            return sum(self.inner.score_self(u) for u in x)
        else:
            return self.score(x, x)


class GraphMatchingMetric(Metric[Graph[T]]):
    """A metric derived from the matching of two graphs (including trees, DAGs, and general graphs)."""
    def __init__(self, inner: Metric[T], constraint: Union[str, MatchingConstraint] = MatchingConstraint.ONE_TO_ONE):
        self.inner = inner
        self.constraint = MatchingConstraint.from_str(constraint) if isinstance(constraint, str) else constraint

    def score(self, x: Graph[T], y: Graph[T]) -> float:
        x_nodes = list(x.nodes())
        y_nodes = list(y.nodes())
        gram_matrix = self.inner.gram_matrix(x_nodes, y_nodes)
        x_reach = _reachability_matrix(x)
        y_reach = _reachability_matrix(y)

        problem = MatchingProblem(x_nodes, y_nodes, gram_matrix, has_vars=False)
        problem.add_matching_constraint(self.constraint)
        problem.add_monotonicity_constraint(x_reach, y_reach)
        return problem.solve()


class LatentSetMatchingMetric(Metric[Collection[T]]):
    """A metric derived to support matching latent variables defined in structures."""

    def __init__(
            self,
            cls: Type[T],
            inner: Metric[T],
            constraint: Union[str, MatchingConstraint] = MatchingConstraint.ONE_TO_ONE
    ):
        if is_dataclass(cls):
            self.cls = cls
        else:
            raise ValueError(f"{cls} has to be a dataclass.")
        self.inner = inner
        self.constraint = MatchingConstraint.from_str(constraint) if isinstance(constraint, str) else constraint

    def score(self, x: Collection[T], y: Collection[T]) -> float:
        """Score two collections of objects."""
        x = list(x)
        y = list(y)
        gram_matrix = self.inner.gram_matrix(x, y)
        problem = MatchingProblem(x, y, gram_matrix, has_vars=True)
        problem.add_matching_constraint(self.constraint)
        problem.add_variable_matching_constraint()
        problem.add_latent_variable_constraint(self.cls)
        return problem.solve()

    def score_self(self, x: Collection[T]) -> float:
        """Score a collection of objects with itself."""
        return SetMatchingMetric(self.inner, self.constraint).score_self(x)




"""Normalizers to normalize metrics as normalized metrics."""
from typing import Optional, Protocol, TypeVar, runtime_checkable

from metametric.core.metric import Metric

T = TypeVar("T")


@runtime_checkable
class Normalizer(Protocol):
    """A metric that normalizes another metric."""

    def normalize(self, score_xy: float, score_xx: float, score_yy: float) -> float:
        """Normalize the metric.

        Args:
            score_xy (`float`): The score between two objects.
            score_xx (`float`): The score of the first object with itself, usually on the prediction side.
            score_yy (`float`): The score of the second object with itself, usually on the reference side.

        Returns:
            `float`: The normalized score.
        """
        raise NotImplementedError()

    @property
    def name(self) -> str:
        """Get the name of the normalizer."""
        raise NotImplementedError()

    @staticmethod
    def from_str(s: str) -> Optional["Normalizer"]:
        if s == "none":
            return None
        if s == "jaccard":
            return Jaccard()
        elif s == "precision":
            return Precision()
        elif s == "recall":
            return Recall()
        elif s == "dice":
            return FScore()
        elif s.startswith("f"):
            return FScore(beta=float(s[1:]))
        else:
            raise ValueError(f"Unknown normalizer {s}")


class Jaccard(Normalizer):
    """Jaccard metric."""

    def normalize(self, score_xy: float, score_xx: float, score_yy: float) -> float:
        """Normalize the metric using Jaccard metric."""
        return score_xy / (score_xx + score_yy - score_xy)

    @property
    def name(self) -> str:
        """Get the name of the normalizer."""
        return "jaccard"


class Precision(Normalizer):
    """Precision metric."""

    def normalize(self, score_xy: float, score_xx: float, score_yy: float) -> float:
        """Normalize the metric using precision metric."""
        return score_xy / score_xx

    @property
    def name(self) -> str:
        """Get the name of the normalizer."""
        return "precision"


class Recall(Normalizer):
    """Recall metric."""

    def normalize(self, score_xy: float, score_xx: float, score_yy: float) -> float:
        """Normalize the metric using recall metric."""
        return score_xy / score_yy

    @property
    def name(self) -> str:
        """Get the name of the normalizer."""
        return "recall"


class FScore(Normalizer):
    """F-score metric."""

    def __init__(self, beta: float = 1.0):
        self.beta = beta

    def normalize(self, score_xy: float, score_xx: float, score_yy: float) -> float:
        """Normalize the metric using F-score metric."""
        return (1 + self.beta**2) * score_xy / ((self.beta**2) * score_yy + score_xx) if score_xy > 0.0 else 0.0

    @property
    def name(self) -> str:
        """Get the name of the FScore based on `beta`."""
        if self.beta == 1.0:
            return "f1"
        else:
            b = int(self.beta) if self.beta.is_integer() else self.beta
            return f"f{b}"


class NormalizedMetric(Metric[T]):
    """A wrapper for the metric that normalizes another metric.

    This ensures that applying a [`Normalizer`] to a [`Metric`] is also a [`Metric`].
    """

    def __init__(self, inner: Metric[T], normalizer: Normalizer):
        self.inner = inner
        self.normalizer = normalizer

    def score(self, x: T, y: T) -> float:
        """Score two objects."""
        sxy = self.inner.score(x, y)
        sxx = self.inner.score_self(x)
        syy = self.inner.score_self(y)
        return self.normalizer.normalize(sxy, sxx, syy)

    def score_self(self, x: T) -> float:
        """Score an object with itself."""
        return 1.0



"""Metric aggregator for computing metrics on a batch of predictions and references."""
from typing import Callable, Collection, Dict, Optional, Protocol

from metametric.core.normalizers import Normalizer
from metametric.core.state import SingleMetricState


def _compute_normalized_metrics(
        normalizers: Collection[Optional[Normalizer]],
        sxy: float,
        sxx: float,
        syy: float
) -> Dict[str, float]:
    normalized_metrics = {
        normalizer.name: normalizer.normalize(sxy, sxx, syy)
        for normalizer in normalizers if normalizer is not None
    }
    if None in normalizers:
        normalized_metrics[""] = sxy
    return normalized_metrics


class Reduction(Protocol):
    """Describes how a collection of metric results are reduced.

    Examples include macro-averaging and micro-averaging.
    """

    def compute(self, agg: SingleMetricState) -> Dict[str, float]:
        """Compute the metrics from the aggregator."""
        raise NotImplementedError()

    def with_extra(self, extra: Callable[[Dict[str, float]], Dict[str, float]]):
        return ReductionWithExtra(self, extra)


class MacroAverage(Reduction):
    """Macro-average reduction."""
    def __init__(self, normalizers: Collection[Optional[Normalizer]]):
        self.normalizers = normalizers
        self.normalizer_names = [normalizer.name for normalizer in normalizers if normalizer is not None]
        if None in normalizers:
            self.normalizer_names.append("")

    def compute(self, state: SingleMetricState) -> Dict[str, float]:
        n = len(state)
        metrics_per_sample = [
            _compute_normalized_metrics(self.normalizers, sxy, sxx, syy)
            for sxy, sxx, syy in zip(state.matches, state.preds, state.refs)
        ]
        metrics = {
            name: sum(metric[name] for metric in metrics_per_sample) / n
            for name in self.normalizer_names
        }
        return metrics


class MicroAverage(Reduction):
    """Micro-average reduction."""
    def __init__(self, normalizers: Collection[Optional[Normalizer]]):
        self.normalizers = normalizers

    def compute(self, state: SingleMetricState) -> Dict[str, float]:
        sxy_total = sum(state.matches)
        sxx_total = sum(state.preds)
        syy_total = sum(state.refs)
        metrics = {
            name: value
            for name, value in _compute_normalized_metrics(self.normalizers, sxy_total, sxx_total, syy_total).items()
        }
        return metrics


class MultipleReductions(Reduction):
    """A collection of multiple reductions."""
    def __init__(self, reductions: Dict[str, Reduction]):
        self.reductions = reductions

    def compute(self, state: SingleMetricState) -> Dict[str, float]:
        return {
            (f"{prefix}-{name}" if name != "" else prefix): value
            for prefix, family in self.reductions.items()
            for name, value in family.compute(state).items()
        }


class ReductionWithExtra(Reduction):
    """Equip a downstream function after reduction is computed."""
    def __init__(self, original: Reduction, extra: Callable[[Dict[str, float]], Dict[str, float]]):
        self.original = original
        self.extra = extra

    def compute(self, state: SingleMetricState) -> Dict[str, float]:
        metrics = self.original.compute(state)
        metrics.update(self.extra(metrics))
        return metrics



"""Metric interface and implementations for commonly used metrics."""
from abc import abstractmethod
from dataclasses import dataclass, is_dataclass
from functools import reduce
from operator import mul
from typing import (Callable, ClassVar, Collection, Dict, Generic, Protocol,
                    Type, TypeVar, Union, get_origin, runtime_checkable)

import numpy as np

S = TypeVar("S", contravariant=True)
T = TypeVar("T", contravariant=True)
U = TypeVar("U")


class Metric(Generic[T]):
    """Metric interface."""

    @abstractmethod
    def score(self, x: T, y: T) -> float:
        """Score two objects."""
        raise NotImplementedError()

    def score_self(self, x: T) -> float:
        """Scores an object against itself.

        In many cases there is a faster way to compute this than the general pair case.
        In such cases, please override this function.
        """
        return self.score(x, x)

    def gram_matrix(self, xs: Collection[T], ys: Collection[T]) -> np.ndarray:
        """Compute the gram matrix of the metric."""
        return np.array([[self.score(x, y) for y in ys] for x in xs])

    def contramap(self, f: Callable[[S], T]) -> "Metric[S]":
        """Returns a new metric by first preprocess the objects by a given function."""
        return ContramappedMetric(self, f)

    @staticmethod
    def from_function(f: Callable[[T, T], float]) -> "Metric[T]":
        """Create a metric from a function.

        Args:
            f (`Callable[[T, T], float]`):
                A function that takes two objects and returns a float.
                This is the function that derives the metric.

        Returns:
            `Metric[T]`: A metric that uses the function to score two objects.
        """
        return MetricFromFunction(f)


class MetricFromFunction(Metric[T]):
    """A metric wrapped from a function."""

    def __init__(self, f: Callable[[T, T], float]):
        self.f = f

    def score(self, x: T, y: T) -> float:
        """Score two objects."""
        return self.f(x, y)


class ContramappedMetric(Metric[S]):
    """A metric contramapped by a function."""

    def __init__(self, inner: Metric[T], f: Callable[[S], T]):
        self.inner = inner
        self.f = f

    def score(self, x: S, y: S) -> float:
        """Score two objects."""
        return self.inner.score(self.f(x), self.f(y))

    def score_self(self, x: S) -> float:
        """Scores an object against itself."""
        return self.inner.score_self(self.f(x))


class DiscreteMetric(Metric[T]):
    """A metric for discrete objects."""

    def __init__(self, cls: Type[T]):
        if getattr(cls, "__eq__", None) is None:
            raise ValueError("Class must implement __eq__")

    def score(self, x: T, y: T) -> float:
        """Score two objects."""
        return float(x == y)

    def score_self(self, x: T) -> float:
        """Scores an object against itself."""
        return 1.0


class ProductMetric(Metric[T]):
    """A metric that is the product of other metrics."""

    def __init__(self, cls: Type[T], field_metrics: Dict[str, Metric]):
        if not is_dataclass(cls):
            raise ValueError(f"{cls} has to be a dataclass.")
        self.field_metrics = field_metrics

    def score(self, x: T, y: T) -> float:
        """Score two objects."""
        return reduce(
            mul, (self.field_metrics[fld].score(getattr(x, fld), getattr(y, fld)) for fld in self.field_metrics.keys())
        )


class UnionMetric(Metric[T]):
    """A metric that is the union of other metrics."""

    def __init__(self, cls: Type[T], case_metrics: Dict[type, Metric]):
        if get_origin(cls) is not Union:
            raise ValueError(f"{cls} has to be a union.")
        self.case_metrics = case_metrics

    def score(self, x: T, y: T) -> float:
        """Score two objects."""
        x_type = type(x)
        y_type = type(y)
        if x_type != y_type:
            return 0.0
        return self.case_metrics[x_type].score(x, y)

    def score_self(self, x: T) -> float:
        """Scores an object against itself."""
        return 1.0


@dataclass(eq=True, frozen=True)
class Variable:
    """A variable in latent matchings."""

    name: str

    latent_metric: ClassVar[Metric["Variable"]] = Metric.from_function(lambda x, y: 1.0)


@runtime_checkable
class HasMetric(Protocol[T]):
    """Protocol for classes that have a metric."""

    metric: Metric[T]


@runtime_checkable
class HasLatentMetric(Protocol[T]):
    """Protocol for classes that have a latent metric."""

    latent_metric: Metric[T]
